/* --------------------------------------------------------------
    Signed integers with unlimited range.
    Input-output.

    http://www.imach.uran.ru/cbignum

    Copyright 1999-2003 by Raul N.Shakirov, IMach of RAS(UB).
    All Rights Reserved.

    Permission has been granted to copy, distribute and modify
    software in any context without fee, including a commercial
    application, provided that the aforesaid copyright statement
    is present here as well as exhaustive description of changes.

    THE SOFTWARE IS DISTRIBUTED "AS IS". NO WARRANTY OF ANY KIND
    IS EXPRESSED OR IMPLIED. YOU USE AT YOUR OWN RISK. THE AUTHOR
    WILL NOT BE LIABLE FOR DATA LOSS, DAMAGES, LOSS OF PROFITS OR
    ANY OTHER KIND OF LOSS WHILE USING OR MISUSING THIS SOFTWARE.
-------------------------------------------------------------- */
#ifdef  _MSC_VER
#pragma warning(disable:4995)
#endif/*_MSC_VER*/
#if __TURBOC__ > 0x410
#pragma option -xd-
#endif
#include <assert.h>
#include <ctype.h>
#include <stdlib.h>
#include <iomanip.h>
#include <limits.h>
#include "Cbignum.h"

#define MAX_RADIX 16

#define MIN_PROGR 10

static const char numtochar [2] [MAX_RADIX + 1] =
                               {{'0','1','2','3','4','5','6','7','8','9',
                                 'a','b','c','d','e','f','x'},
                                {'0','1','2','3','4','5','6','7','8','9',
                                 'A','B','C','D','E','F','X'}};

static char chartonum [UCHAR_MAX + 1] = { 0 };

static const long maxradix [] [4] = {{ 0L,  1},
                                     { 1L,  1},
                            { 1073741824L, 30},
                            { 1162261467L, 19},
                            { 1073741824L, 15},
                            { 1220703125L, 13},
                            {  362797056L, 11},
                            { 1977326743L, 11},
                            { 1073741824L, 10},
                            {  387420489L,  9},
                            { 1000000000L,  9},
                            {  214358881L,  8},
                            {  429981696L,  8},
                            {  815730721L,  8},
                            { 1475789056L,  8},
                            {  170859375L,  7},
                            {  268435456L,  7} };

#define MAX_NP_IN  ((sizeof(size_t))==2? 2: 9)
#define MAX_NP_OUT ((sizeof(size_t))==2? 2: 9)
#define MAX_NP    (MAX_NP_IN>MAX_NP_OUT?MAX_NP_IN:MAX_NP_OUT)

static exarray<cBigNumber> powradix [MAX_RADIX + 1];

static  size_t  cBigNumberSqrTab (
                unsigned radix,
                size_t len,
                size_t max_np
)
{
  assert (radix >= 2);
  assert (radix <= MAX_RADIX);
  assert (max_np <= MAX_NP);
  static size_t npow [MAX_RADIX + 1];

  if (npow [radix] == 0)
  {
    int i;
    for (i = 0; i <= UCHAR_MAX; i++)
    {
      chartonum [i] = MAX_RADIX;
    }
    for (i = 0; i < MAX_RADIX; i++)
    {
      chartonum [numtochar [0] [i]] = (char)i;
      chartonum [numtochar [1] [i]] = (char)i;
    }

    powradix [radix] [0] = maxradix [radix] [0];
    powradix [radix] [0].tab();
    powradix [radix] [1] = maxradix [radix] [0];
    powradix [radix] [1].pow2();
    powradix [radix] [1].tab();
    npow [radix] = 2;
  }

  size_t np = 0;
  while (np < max_np && powradix [radix] [np].length() * 4 < len)
  {
    if (++np >= npow [radix])
    {
      powradix [radix] [np] = powradix [radix] [np - 1];
      powradix [radix] [np].pow2();
      powradix [radix] [np].tab();
      ++ npow [radix];
    }
  }
  assert (np <= max_np);
  return np;
}

static cBigNumber           num_;
static cBigNumber           mod_;
static exvector<cBigNumber> stk_(MAX_NP);
static exvector<int>       kstk_(MAX_NP);

cBigNumber& cBigNumber::set_ (
                const EXPTR(char) ps,
                unsigned radix,
                unsigned len
            )
{
  assert (ps);
  assert (radix >= 2);
  assert (radix <= MAX_RADIX);

  *this = 0;
  if (len != 0)
  {
    long lradix =       maxradix [radix] [0];
    int  nradix = (int) maxradix [radix] [1];
    int nr = (len - 1) / nradix;
    int n = len - nr++ * nradix;

    size_t np = cBigNumberSqrTab (radix, nr, MAX_NP_IN);
    EXPTR(cBigNumber) powr = EXPTRTYPE(powradix [radix]);

    EXPTR(cBigNumber)  mul = EXPTRTYPE(stk_);
    { for (size_t i = 1; i < np; i++) mul [i] = 0; }

    int numlength = nr / (1024 / sizeof (long));
    if (numlength < MIN_PROGR) numlength = 0;
    if (numlength > 0)
    {
      cerr << "Converting"
           << setw (7) << 0         <<"K of"
           << setw (7) << numlength <<"K\b"
           << "\b\b\b\b\b\b\b\b\b\b\b\b\b\b"
           << "\b\b\b\b\b\b\b\b\b\b\b\b\b\b";
    }

    long lhigh = 0;
    if ((nr & 1) != 0) goto odd;
    do
    {
      {
        long lb = 0;
        for(;;)
        {
          lb += chartonum [*ps++];
          assert (lb >= 0);
          if (--n == 0) break;
          lb *= radix;
        }
        n = nradix;
        lhigh = lb;
      }
      --nr;

odd:  {
        long lb = 0;
        for(;;)
        {
          lb += chartonum [*ps++];
          assert (lb >= 0);
          if (--n == 0) break;
          lb *= radix;
        }
        n = nradix;
        num_ = lb;
      }
      --nr;

      num_.addmul (lhigh, lradix);
      assert (num_ >= 0);

      for (size_t i = 1;; i++)
      {
        if (i < np)
        {
           if ((nr & (2<<i)-1) != 0)
           {
             num_.swap (mul [i]);
             break;
           }
           num_.addmultab (powr [i], mul [i]);
           assert (num_ >= 0);
        }
        else
        {
          if (numlength > 0)
          {
            cerr << "Converting"
                 << setw (7) << (numlength -
                                 nr / (1024 / sizeof (long)))
                                           <<"K of"
                 << setw (7) <<  numlength <<"K\b"
                 << "\b\b\b\b\b\b\b\b\b\b\b\b\b\b"
                 << "\b\b\b\b\b\b\b\b\b\b\b\b\b\b";
          }
          num_.addmultab (powr [i], *this);
          assert (num_ >= 0);
          num_.swap (*this);
          break;
        }
      }
    }
    while (nr != 0);

    if (numlength != 0)
          cerr << "                             \b"
               << "\b\b\b\b\b\b\b\b\b\b\b\b\b\b"
               << "\b\b\b\b\b\b\b\b\b\b\b\b\b\b";
  }

  EXDEBUG (num_ = 0);
  return *this;
}

cBigNumber& cBigNumber::set (
                const char *psz,
                unsigned radix
            )
{
  assert (psz);
  assert (radix != 1);
  assert (radix <= MAX_RADIX);
  int sign = 0;
  unsigned c = *psz++;

  switch (c)
  {
    case '-': sign = 1;
    case '+': c = *psz++;
  }

  if (radix == 0)
  {
    radix = 10;
    if (c == '0')
    {
      c = *psz++; radix = 8;
      if (c == 'X' || c == 'x') { c = *psz++; radix = 16; }
    }
  }

  cBigNumberSqrTab (radix, 0, 0);
  int len = 0;
  for (;;)
  {
    if ((size_t)chartonum [c] >= radix) break;
    c = psz [len++];
  }
  --psz;

  set_ (EXPTRTO(char,psz,len), radix, len);

  if (sign) this->neg();

  return *this;
}

istream& operator >> (istream& is, cBigNumber& b)
{
  assert (&is);
  assert (&b);
  static exarray<char> str_;
  int sign = 0;
  char c;
  long flags = is.flags();
  unsigned radix = (
         (flags & ios::dec)? 10:
         (flags & ios::oct)? 8:
         (flags & ios::hex)? 16: 0);

  is >> c;

  switch (c)
  {
    case '-': sign = 1;
    case '+': c = (char) is.get();
  }
  if (radix == 0)
  {
    radix = 10;
    if (c == '0')
    {
      c = (char) is.get(); radix = 8;
      if (c == 'X' || c == 'x') { c = (char) is.get(); radix = 16; }
    }
  }

  cBigNumberSqrTab (radix, 0, 0);
  int len = 0;
  for (;;)
  {
    if ((size_t)chartonum [(unsigned char)c] >= radix) break;
    str_ [len++] = c;
    c = (char) is.get();
  }
  if ((signed char)c != EOF)
  {
    is.putback (c);
  }
  str_ [len] = 0;
  str_.checkshrink (len);

  b.set_ (EXPTRTYPE(str_), radix, len);

  if (sign) b.neg();

  return is;
}

char*   cBigNumber::toatmp (
                unsigned radix,
                ios& os
        ) const
{
  assert (radix != 1);
  assert (radix <= MAX_RADIX);
  num_ = *this; num_.fit();
  size_t i = 0;
  long flags = (&os? os.flags(): 0);
  static exarray<char> str_;

  register const EXPTR(char) ps = EXPTRTO (char,
                         numtochar [(flags & ios::uppercase) != 0],
                         sizeof (numtochar [0]));

  {
    long sign = num_.comp0();

    if (radix == 0)
    {
      radix = ((flags & ios::hex)? 16: (flags & ios::oct)? 8: 10);
      if (radix != 10)
      {
        if (sign < 0)
        {
          ++(num_.e[0]);
          num_.checkexpand (num_.length());
          num_.e [num_.length()] = 0;
        }
        sign = 0;
      }
    }

    if (sign < 0)
    {
      str_ [i++] = '-'; num_.neg();
    }
    else if (sign > 0 && (flags & ios::showpos))
    {
      str_ [i++] = '+';
    }

    if (radix % 8 == 0 && (flags & ios::showbase))
    {
      str_ [i++] = '0'; if (radix == 16) str_ [i++] = ps [16];
    }
  }

  size_t j = i;
  {
    size_t np = cBigNumberSqrTab (radix, num_.length(), MAX_NP_OUT);
    EXPTR(cBigNumber) powr = EXPTRTYPE(powradix [radix]);

    EXPTR(cBigNumber)  div = EXPTRTYPE(stk_);
    EXPTR(int)        kdiv = EXPTRTYPE(kstk_);

    int numlength = num_.length() / (1024 / sizeof (long));
    if (numlength < MIN_PROGR) numlength = 0;
    size_t npmax = np;
    for (;;)
    {
      if (np == npmax)
      {
        if (numlength > 0)
        {
          cerr << "Formatting"
               << setw (7) << (numlength -
                             num_.length() / (1024 / sizeof (long)))
                                         <<"K of"
               << setw (7) <<  numlength <<"K\b"
               << "\b\b\b\b\b\b\b\b\b\b\b\b\b\b"
               << "\b\b\b\b\b\b\b\b\b\b\b\b\b\b";
        }

        num_.swap (mod_);
        num_.setdivmodtab (mod_, powr [np]);

        if (num_.comp0() == 0)
        {
          if (np == 0) break;
          num_.swap (mod_);
          npmax = --np;
          numlength = -numlength;
          assert (kdiv [np] == 0);
          continue;
        }
      }
      else
      {
        assert (np > 0);
        assert (kdiv [np - 1] == 0);
        if (kdiv [np] == 0)
        {
          ++np; continue;
        }
        mod_.swap (div [np]);
        kdiv [np] = 0;
      }

      for (; np != 0; np--)
      {
        assert (kdiv [np - 1] == 0);
        kdiv [np - 1] = 1;
        div  [np - 1].setdivmodtab (mod_, powr [np - 1]);
      }

      {
        register long lnum;
        lnum = mod_.loword();
        for (;;)
        {
          register long knum = maxradix [radix] [1] - 1;
          do
          {
            str_ [i++] = ps [ (int) (lnum % radix) ];
            lnum /= radix;
          }
          while (--knum != 0);

          assert (lnum >= 0);
          assert (lnum < (long) radix);
          str_ [i++] = ps [(int)(lnum)];

          if (kdiv [0] == 0) break;
          lnum = div [0].loword();
          kdiv [0] = 0;
        }
      }

      np = (npmax != 0);
    }

    {
      long lnum = mod_.loword();
      if (lnum != 0)
      do
      {
        str_ [i++] = ps [(int) (lnum % radix)];
      }
      while ((lnum /= radix) != 0);
    }
    if (numlength != 0)
          cerr << "                             \b"
               << "\b\b\b\b\b\b\b\b\b\b\b\b\b\b"
               << "\b\b\b\b\b\b\b\b\b\b\b\b\b\b";

    if (i == j) str_ [i++] = ps [0];
  }

  {
    assert (i > j);
    if (flags & ios::internal)
    {
      size_t n = os.width();
      char   c = os.fill();
      while (i < n) str_ [i++] = c;
    }
    str_ [i--] = 0;
  }

  {
    register EXPTR(char) ps = EXPTRTYPE(str_);

    while (i > j)
    {
      char c = ps [j]; ps [j] = ps [i]; j++; ps [i] = c; i--;
    }
  }

  EXDEBUG (num_ = 0);
  EXDEBUG (mod_ = 0);
  return ((char *)(const char *)str_);
}

ostream& operator << (ostream& os, const cBigNumber& b)
{
  assert (&os);
  assert (&b);
  os << b.toatmp (0, os);
  return os;
}
